/*
 * vdi_control.c
 *
 * Copyright 1998 Tomas Berndtsson <tomas@nocrew.org>
 * Copyright 1999 Christer Gustavsson <cg@nocrew.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *  
 * Read the file COPYING for more information.
 *
 */

#undef DEBUGLEVEL

#include <dlfcn.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/shm.h>
#include <unistd.h>

#include "event.h"
#include "inits.h"
#include "ovdisis.h"
#include "various.h"

void vdi_v_clrwk(VDI_Workstation *);	/* Used by v_opnwk */

VDIWK wk[MAX_HANDLES];
int wk_open[MAX_HANDLES];
int first=1; /* Used to clear the above array first time */


#define WSID(id) (0x42434400 | id)

/*
** Description
** Try to allocate a new workstation descriptor with a specified id.
** If successfull the address will be returned. Otherwise NULL will be
** returned.
**
** 1998-10-09 CG
*/
static
VDI_Workstation *
wsalloc (int id) {
  int               shmid;
  VDI_Workstation * ws;

  shmid = shmget (WSID (id),
                  sizeof (VDI_Workstation),
                  IPC_CREAT | IPC_EXCL | 0644);

  if (shmid < 0) {
    /*    perror ("ovdisis: wsalloc: shmget"); */
    return NULL;
  }

  ws = (VDI_Workstation *)shmat (shmid, NULL, 0);

  if ((int)ws == -1) {
    perror ("ovdisis: wsalloc: shmat");
    return NULL;
  }

  return ws;
}


/*
** Description
** Deallocate a workstation descriptor.
**
** 1998-10-09 CG
*/
static
void
wsfree (int id)
{
  int shmid;

  shmid = shmget (WSID (id),
                  sizeof (VDI_Workstation),
                  0);

  if (shmid < 0) {
    perror ("ovdisis: wsfree: shmget");
  }

  if (shmctl(shmid, IPC_RMID, NULL) < 0) {
    perror ("ovdisis: wsfree: shmctl");
  }
}


/*
** Description
** Attach to a workstation descriptor. If the workstation doesn't exist,
** NULL is returned.
**
** 1998-10-09 CG
*/
static
VDI_Workstation *
wsattach (int id)
{
  int               shmid;
  VDI_Workstation * ws;

  shmid = shmget (WSID (id),
                  sizeof (VDI_Workstation),
                  0);

  if (shmid < 0) {
    perror ("ovdisis: wsattach: shmget");
    return NULL;
  }

  ws = (VDI_Workstation *)shmat (shmid, NULL, 0);

  if ((int)ws == -1) {
    perror ("ovdisis: wsattach: shmat");
    return NULL;
  }

  return ws;
}


/*
** Description
** Detach from a workstation descriptor.
**
** 1998-10-09 CG
*/
static
void
wsdetach (VDI_Workstation * ws)
{
  if (shmdt ((char *)ws) < 0) {
    perror ("ovdisis: wsdetach: shmdt");
  }
}


/*
** Description
** Find a visual to use and initialize it
**
** 1999-08-29 CG
*/
static
VDI_Visual *
init_visual (void)
{
  void *       handle;
  char         module_name[256];
  VDI_Visual * (*visual_init)(void);
  char *       error;
  char *       visualtype;

  strcpy(module_name, VISUALSDIR);
  strcat(module_name, "/libovdisis_");

  /* Read visual type from environment variable, 
   * if available, otherwise use oFBis.
   */
  visualtype = getenv("OVDISIS_VISUAL");
  if(visualtype)
    strcat(module_name, visualtype);
  else
    strcat(module_name, "ofbis");

  strcat(module_name, ".so");
  handle = dlopen(module_name, RTLD_NOW);

  if(handle == NULL)
  {
    fprintf(stderr, "Couldn't open visual %s\n", module_name);
    exit(1);
  }

  visual_init = dlsym(handle, "init");

  error = dlerror();
  if(error != NULL)
  {
    fprintf(stderr, "%s\n", error);
    dlclose(handle);
    exit(1);
  }

  return visual_init();
}


/*
** Description
** Implementation of the vdi function v_opnwk
**
** 1998-10-13 CG
** 1998-12-07 CG
** 1999-08-29 CG
*/
void
vdi_v_opnwk(VDI_Workstation *vwk)
{
  int               w;
  VDI_Workstation * new_ws = NULL;

  if(first) {
    for(w=0 ; w<MAX_HANDLES ; w++)
      wk_open[w] = WS_NOTOPEN;
    first = 0;
  }

  /* find first free handle */
  w=0;
  while ((w < MAX_HANDLES) && ((new_ws = wsalloc (w)) == NULL)) {
    w++;
  }

  if (w == MAX_HANDLES) {
    vdipb->contrl[VDI_HANDLE] = 0;	/* Could not open workstation */
    EDEBUG("v_opnwk: No handles left to open physical workstation, sorry!\n");
    return;
  }
  
  wk[w].physical = new_ws;

  wk[w].physical->visual = init_visual();
  wk[w].physical->visual->private = VISUAL_OPEN(wk[w].physical);
  
  if (wk[w].physical->visual->private == NULL) {
    wsfree (w);
    vdipb->contrl[VDI_HANDLE] = 0;	/* Could not open workstation */
    EDEBUG("v_opnwk: Error opening visual!\n");
    return;
  }

  vdipb->contrl[VDI_HANDLE] = w + 1;	/* new vdi handle */
  wk_open[w] = WS_PHYSICAL;
  wk[w].vwk = wk[w].physical;	/* Link workstation to physical */

  wk[w].physical->handle = w + 1;	/* used for debugging */
  wk[w].physical->pid = getpid();	/* pid of current process */

  ADEBUG("v_opnwk: Physical workstation, handle %d opened: %p\n", w + 1, wk[w].physical);
  ADEBUG("v_opnwk: FrameBuffer opened: %p\n", wk[w].physical->fb);

  /* Setting default values */

  /* v_opnvwk and vq_extnd values */
  init_workstation(wk[w].physical);

  wk[w].physical->write_mode = MD_REPLACE;
  wk[w].physical->clip.x1 = 0;
  wk[w].physical->clip.y1 = 0;
  wk[w].physical->clip.x2 = wk[w].physical->dev.attr.xres;
  wk[w].physical->clip.y2 = wk[w].physical->dev.attr.yres;

  IDEBUG("v_opnwk: Number of bitplanes: %d\n", wk[w].physical->inq.attr.planes);

  vdi_v_clrwk(wk[w].physical);	/* Clear screen */

  /* Set default things */
  init_cmap(wk[w].physical);
  init_marker(wk[w].physical);
  init_line(wk[w].physical);
  init_fill(wk[w].physical);
  init_device(wk[w].physical);
  init_text(wk[w].physical);

  ADEBUG("v_opnwk: Default palette set\n");

  /* Reset event vectors */
  wk[w].physical->butv = NULL;
  wk[w].physical->curv = NULL;
  wk[w].physical->motv = NULL;
  wk[w].physical->timv = NULL;

  /* Start event handler */
  start_event_handler (wk[w].physical);

  vdipb->contrl[N_PTSOUT] = 6;
  vdipb->contrl[N_INTOUT] = 45;
}


/*
** Description
** Implementation of v_clswk
**
** 1998-10-14 CG
** 1998-12-07 CG
*/
void
vdi_v_clswk (VDI_Workstation *vwk)
{
  int i, w;

  w = vdipb->contrl[VDI_HANDLE] - 1;

  if (wk_open[w] != WS_PHYSICAL) {
    if (wk_open[w])
      EDEBUG("v_clswk: Physical workstation, handle %d not open!", w + 1);
    else
      EDEBUG("v_clswk: Handle %d is not a physical workstation!", w + 1);
    return;
  }

  /* Destroy event handler */
  stop_event_handler ();

  VISUAL_FREE_CMAP(wk[w].physical);
  VISUAL_CLOSE(wk[w].physical);

  ADEBUG("v_clswk: FrameBuffer closed\n");

  /* Free virtual workstations left open */
  /* Will closing a physical workstation
     also close all its virtual ? */
  /* It will here! */
  for (i = 0; i < MAX_HANDLES; i++)
    if (wk_open[i] && (wk[i].physical == wk[w].physical)) {
      wsfree (i);
      wk_open[i] = WS_NOTOPEN;
      ADEBUG("v_clswk: Freed workstation, handle %d\n", i + 1);
    }
  ADEBUG("v_clswk: Freed physical workstation, handle %d\n", w + 1);

  /* Above will also free the actual physical ws */

  /* Free physical VDI_Workstation */
  /*
     free(wk[w].physical);
     ADEBUG("v_clswk: Freed physical workstation, handle %d\n",w+1);
   */

  wk_open[w] = WS_NOTOPEN;

  vdipb->contrl[N_PTSOUT] = 0;
  vdipb->contrl[N_INTOUT] = 0;
}




/* This function does not open a framebuffer,
   but only allocate a new virtual workstation. */
void vdi_v_opnvwk(VDI_Workstation *vwk)
{
  int i, v = 0, w;
  VDI_Workstation * physical;
  VDI_Workstation * virtual = NULL;

  ADEBUG ("ovdisis: vdi_v_opnvwk: entering\n");
  w = vdipb->contrl[VDI_HANDLE] - 1;	/* Physical workstation */

  physical = wsattach (w);

  if (physical == NULL) {
    vdipb->contrl[VDI_HANDLE] = 0;	/* Could not open virtual workstation */
    EDEBUG("v_opnvwk: Handle %d is not a physical workstation!\n", w + 1);
    return;
  }
  /* find first free handle */
  while ((v < MAX_HANDLES) && ((virtual = wsalloc (v)) == NULL)) {
    v++;
  }
    
  /* none free found */
  if (v == MAX_HANDLES) {
    wsdetach (w);
    wsfree (v);
    vdipb->contrl[VDI_HANDLE] = 0;	/* Could not open virtual workstation */
    EDEBUG("v_opnvwk: We're all out of handles, I'm afraid!\n");
    return;
  }

  wk[v].vwk = virtual;

  ADEBUG("v_opnvwk: New virtual workstation, handle %d opened linked to %d\n", v + 1, w + 1);

  vdipb->contrl[VDI_HANDLE] = v + 1;	/* new vdi handle */
  wk_open[v] = WS_VIRTUAL;
  wk[v].physical = physical;	/* link virtual to physical, sort of */

  wk[v].vwk->handle = v + 1;	/* used for debugging */
  wk[v].vwk->pid = getpid();	/* pid of current process */

  /* If the virtual workstation is opened from a different
   * process than the physical, we need to open the
   * frame buffer again, using the same console.
   */
#if 0 /* FIXME */
  if(wk[v].vwk->pid != wk[v].physical->pid) {
    if ((wk[v].vwk->fb = FBopen(NULL, FB_KEEP_CURRENT_VC /* | FB_NO_KBD */)) == NULL) {
      wsdetach (w);
      wsfree (v);
      vdipb->contrl[VDI_HANDLE] = 0;	/* Could not open workstation */
      EDEBUG("v_opnvwk: Error opening FrameBuffer!\n");
      return;
    }
  } else { /* Otherwise we use the same framebuffer */
#endif
    wk[v].vwk->visual = wk[v].physical->visual;
#if 0
  }
#endif

  ADEBUG ("ovdisis: vdi_control.c: fb = 0x%x\n", wk[v].vwk->fb);

  /* Copy some things */
  copy_workstation(wk[v].physical, wk[v].vwk);
  copy_cmap(wk[v].physical, wk[v].vwk);
  copy_marker(wk[v].physical, wk[v].vwk);
  copy_line(wk[v].physical, wk[v].vwk);
  copy_fill(wk[v].physical, wk[v].vwk);
  copy_device(wk[v].physical, wk[v].vwk);
  copy_text(wk[v].physical, wk[v].vwk);

  /* make the changes the user wants */
  /* intin[0] - workstation (1-10 == screen) */
  wk[v].vwk->line_a.type = (int)vdipb->intin[1];
  wk[v].vwk->line_a.color = (int)vdipb->intin[2];
  wk[v].vwk->marker_a.type = (int)vdipb->intin[3];
  wk[v].vwk->marker_a.color = (int)vdipb->intin[4];
  wk[v].vwk->text_a.fontid = (int)vdipb->intin[5];
  wk[v].vwk->text_a.color = (int)vdipb->intin[6];
  wk[v].vwk->fill_a.interior = (int)vdipb->intin[7];
  wk[v].vwk->fill_a.style = (int)vdipb->intin[8];
  wk[v].vwk->fill_a.color = (int)vdipb->intin[9];
  /* intin[10] - Coordinate system.  Only used in GDOS */

  wk[v].vwk->write_mode = wk[v].physical->write_mode;
  wk[v].vwk->clip.x1 = wk[v].physical->clip.x1;
  wk[v].vwk->clip.y1 = wk[v].physical->clip.y1;
  wk[v].vwk->clip.x2 = wk[v].physical->clip.x2;
  wk[v].vwk->clip.y2 = wk[v].physical->clip.y2;

  /* v_opnvwk values to return */
  for (i = 0; i < 45; i++)
    vdipb->intout[i] = wk[v].vwk->dev.tab[i] = wk[v].physical->dev.tab[i];
  for (i = 0; i < 12; i++)
    vdipb->ptsout[i] = wk[v].vwk->dev.tab[i + 45] = wk[v].physical->dev.tab[i + 45];

  vdipb->contrl[N_PTSOUT] = 6;
  vdipb->contrl[N_INTOUT] = 45;

  ADEBUG ("ovdisis: vdi_control.c: fb = 0x%x\n", wk[v].vwk->fb);
}

void vdi_v_clsvwk(VDI_Workstation *vwk)
{
  int v;

  v = vdipb->contrl[VDI_HANDLE] - 1;

  if (wk_open[v] != WS_VIRTUAL) {
    if (wk_open[v])
      EDEBUG("v_clsvwk: Handle %d is not a virtual workstation\n", v + 1);
    else
      EDEBUG("v_clsvwk: Handle %d is not open\n", v + 1);
    return;
  }

  /* If we're not in the same process as where the physical
   * workstation was opened, we need to close the framebuffer
   * we opened in v_opnvwk.
   */
  if(wk[v].vwk->pid != wk[v].physical->pid) {
    VISUAL_FREE_CMAP(wk[v].vwk);

#if 0 /* FIXME */
    FBclose(wk[v].vwk->fb);
    ADEBUG("v_clswk: FrameBuffer closed\n");  
#endif

  }

  wsfree (v);
  wk_open[v] = WS_NOTOPEN;

  vdipb->contrl[N_PTSOUT] = 0;
  vdipb->contrl[N_INTOUT] = 0;

  ADEBUG("v_clsvwk: Virtual workstation, handle %d freed\n", v + 1);
}

void vdi_v_clrwk(VDI_Workstation *vwk)
{
  VISUAL_CLEAR(vwk);
  vdipb->contrl[N_PTSOUT] = 0;
  vdipb->contrl[N_INTOUT] = 0;
}

void vdi_v_updwk(VDI_Workstation *vwk)
{
  EDEBUG("v_updwk: Call not implemented!\n");
  vdipb->contrl[N_PTSOUT] = 0;
  vdipb->contrl[N_INTOUT] = 0;
}

void vdi_vs_clip(VDI_Workstation *vwk)
{
  if (vdipb->intin[0]) {	/* Clipping on */
    vwk->clip.x1 = vdipb->ptsin[0];
    vwk->clip.y1 = vdipb->ptsin[1];
    vwk->clip.x2 = vdipb->ptsin[2];
    vwk->clip.y2 = vdipb->ptsin[3];
    fix_rect(&vwk->clip);
    vwk->clip.x1 = max(0, vdipb->ptsin[0]);
    vwk->clip.y1 = max(0, vdipb->ptsin[1]);
    vwk->clip.x2 = min(vwk->dev.attr.xres, vdipb->ptsin[2]);
    vwk->clip.y2 = min(vwk->dev.attr.yres, vdipb->ptsin[3]);
  } else {			/* Clipping off */
    vwk->clip.x1 = 0;
    vwk->clip.y1 = 0;
    vwk->clip.x2 = vwk->dev.attr.xres;
    vwk->clip.y2 = vwk->dev.attr.yres;
  }
  ADEBUG("vs_clip: Clipping for handle %d set to: %d, %d, %d, %d.\n",
    vwk->handle, vwk->clip.x1, vwk->clip.y1, vwk->clip.x2, vwk->clip.y2);

  vdipb->contrl[N_PTSOUT] = 0;
  vdipb->contrl[N_INTOUT] = 0;
}


/*
** Exported
**
** 1998-12-26 CG
** 1999-05-22 CG
*/
void
vdi_vswr_mode (VDI_Workstation * vwk) {
  VISUAL_SET_WRITE_MODE(vwk, vdipb->intin[0]);

  vwk->write_mode = vdipb->intin[0];
  vdipb->intout[0] = vwk->write_mode;

  vdipb->contrl[N_PTSOUT] = 0;
  vdipb->contrl[N_INTOUT] = 1;
}

void vdi_vq_extnd(VDI_Workstation *vwk)
{
  int i;

  if (vdipb->intin[0]) {	/* extended answers */
    for (i = 0; i < 45; i++)
      vdipb->intout[i] = vwk->inq.tab[i];
    for (i = 0; i < 12; i++)
      vdipb->ptsout[i] = vwk->inq.tab[i + 45];
  } else {			/* same answers as from v_opnvwk */
    for (i = 0; i < 45; i++)
      vdipb->intout[i] = vwk->dev.tab[i];
    for (i = 0; i < 12; i++)
      vdipb->ptsout[i] = vwk->dev.tab[i + 45];
  }

  vdipb->contrl[N_PTSOUT] = 6;
  vdipb->contrl[N_INTOUT] = 45;
}
